#include <stdio.h>
#include <stdlib.h>

#include "btree.h"
 
btree64 * create_head (/*btree * head,*/ /*int num_nodes, unsigned int * keys*/ TCHAR * buffer, int num_keys, char flag)
{
	btree64 * head;
	BOOL err;

	head = /*(btree64*) HeapAlloc(GetProcessHeap(),HEAP_ZERO_MEMORY,sizeof(btree64));*/malloc(sizeof(btree));
	if ( head == NULL)
		return NULL;
	else
	{	
		err = populate1(buffer,head,num_keys,flag);
		if ( err == FALSE)
			return FALSE;
		else
			return head;
	}
}

btree64 * add_node(TCHAR * buffer, int num_keys, char flag)
{
	btree64 * node;
	BOOL err;

	node = /*(btree64*) HeapAlloc(GetProcessHeap(),HEAP_ZERO_MEMORY,sizeof(btree64));*/malloc(sizeof(btree));
	if ( node == NULL)
		return NULL;
	else
	{	
		err = populate1(buffer,node,num_keys,flag);
		if ( err == FALSE)
			return FALSE;
		else
			return node;
	}
}

tuple64* populate_keys(unsigned __int64 * keys, unsigned __int64 * backpointers, unsigned __int64 * offsets, int num_nodes , int num_keys)
{
	tuple64 * t_keys;
	int i = 0;

	t_keys = /*(tuple64 *) HeapAlloc(GetProcessHeap(),HEAP_ZERO_MEMORY,sizeof(tuple64));*/(tuple64 *) malloc(sizeof(tuple64) * num_nodes);
	if (t_keys == NULL)
		return NULL;	

	for(i = 0; i < num_nodes; i++)
	{
		t_keys[i].i_id = keys[i];
		t_keys[i].backpointer = backpointers[i];
		t_keys[i].offset = offsets[i];
	}

	return t_keys;
}

BOOL populate1(TCHAR * buffer, btree64 * node, int num_keys,char flag)
{
	unsigned char items;              //offset 0x1f0, size 1 byte
	unsigned char max_items;          //offset 0x1f1, size 1 byte  
	unsigned char item_size;          //offset 0x1f2, size 1 byte  
	unsigned char node_level;         //offset 0x1f3, size 1 byte
	unsigned short type_indicator;    //offset 0x1f4, size 2 bytes
	int offset = (flag == 1 ? 0x1f0 : 0x1e8);
	BOOL err;

	items = (unsigned char) buffer[offset++];
	max_items = (unsigned char) buffer[offset++];
	item_size = (unsigned char) buffer[offset++];
	node_level = (unsigned char) buffer[offset++];
	memcpy(&type_indicator,&buffer[offset],2);

	node->is_leaf = node_level;
	if(type_indicator == 0x8080)
		node->node_type = 1;
	else if(type_indicator == 0x8181)
		node->node_type = 2;
	node->next = NULL;
	node->num_nodes = items;
	if ( ((node->is_leaf != 0 || (node->is_leaf == 0 && max_items != 0x1f) ) && flag == 1) ||
	   (node->is_leaf != 0 || (node->is_leaf == 0 && max_items != 0x0f) && flag == 2))
	{
		err = populate2(buffer,node,items,item_size,num_keys,flag);
		if (err == FALSE)
			return FALSE;
	}
	else
	{
		err = populate3(buffer,node,items,item_size,flag);
	}

	return TRUE;
}

BOOL populate2(TCHAR * buffer, btree64 * node, unsigned int items, unsigned int item_size, int num_keys, char flag)
{
	unsigned __int64 * keys;
	unsigned __int64 * backpointers;
	unsigned __int64 * offsets;
	unsigned int i = 0, j = 0;
	int step = (flag == 1 ? 4 : 8);

	keys = (unsigned __int64 *) malloc(sizeof(unsigned __int64) * items);
	if(keys == NULL)
		return FALSE;
	backpointers = (unsigned __int64 *) malloc(sizeof(unsigned __int64) * items);
	if(backpointers == NULL)
		return FALSE;
	offsets = (unsigned __int64 *) malloc(sizeof(unsigned __int64) * items);
	if(offsets == NULL)
		return FALSE;

	while(i < (items * item_size) && j < items)
	{
		keys[j] &= 0;
		memcpy(&keys[j],&buffer[i],step);
		i += step;
		backpointers[j] &= 0;
		memcpy(&backpointers[j],&buffer[i],step);
		i += step;
		offsets[j] &= 0;
		memcpy(&offsets[j],&buffer[i],step);
		i += step;
		j ++;
	}

	node->key = populate_keys(keys,backpointers,offsets,items,num_keys);

	free(keys);
	free(backpointers);
	free(offsets);

	return TRUE;
}

BOOL populate3(TCHAR * buffer, btree64 * node, unsigned int items, unsigned int item_size, char flag)
{
	unsigned int i = 0,j = 0;
	int step = (flag == 1 ? 4 : 8);

	node->leaf_d = (tuple_leaf_d64 *) malloc(sizeof(tuple_leaf_d64) * items);
	if (node->leaf_d == NULL)
		return FALSE;

	while ( i <  (items * item_size) && j < items)
	{
		node->leaf_d[j].D_ID = 0;
		memcpy(&node->leaf_d[j].D_ID,&buffer[i],step);
		i += step;
		node->leaf_d[j].DESC_I_ID = 0;
		memcpy(&node->leaf_d[j].DESC_I_ID,&buffer[i],step);
		i += step;
		node->leaf_d[j].TREE_I_ID = 0;
		memcpy(&node->leaf_d[j].TREE_I_ID,&buffer[i],step);
		i += step;
		node->leaf_d[j].PARENT_D_ID = 0;
		memcpy(&node->leaf_d[j].PARENT_D_ID,&buffer[i],4);
		i += step;
		j++;
	}

	return TRUE;
}

void release_bbt(btree64 * head)
{
	int i;

	if(head->is_leaf == 0)
	{
		if(head->node_type = 0x01)
			free(head->leaf_d);
		else
			free(head->key);
		free(head);
	}
	else
	{
		for(i = 0; i< head->num_nodes; i++)
		{
			release_bbt(head->next[i]);
		}
		free(head->key);
		free(head);
	}
}

btree64 * get_node (btree64 * tree, unsigned __int64 i_id, int * index)
{
	int i;
	btree64 * node = tree;

	if(node->is_leaf > 0)
	{
		for(i = 0; i < node->num_nodes - 1; i++)
		{
			if(node->key[i].i_id <= i_id && node->key[i + 1].i_id > i_id)
				return get_node(node->next[i],i_id,index);
		}

		return get_node(node->next[i],i_id,index);
	}
	else
	{
		for(i = 0; i < node->num_nodes - 1; i++)
		{
			if(node->leaf_d[i].D_ID == i_id)
			{
				*index = i; 
				return node;
			}
		}
	}

	return node;	
}