#ifndef _BTREE_H_
#define _BTREE_H_

#include <tchar.h>
#include <windows.h>
#include <malloc.h>

typedef struct _tuple{
	//tuples for btree node
	unsigned int i_id;
	unsigned int backpointer;
	unsigned int offset;
}tuple;

//descriptor leaf has different structure then other nodes.
typedef struct _tuple_leaf_d{
	unsigned int D_ID;
	unsigned int DESC_I_ID;
	unsigned int TREE_I_ID;
	unsigned int PARENT_D_ID;
}tuple_leaf_d;

typedef struct _btree{
	int num_nodes;         //number of keys in the current node.
	union /*nodes*/
	{
		tuple * key;          //array of structs representing all keys in a given btree
		tuple_leaf_d * leaf_d;
	};
	int is_leaf;           //int if 0 means this node is a leaf, otherwise if internal node, node with highest index is root.
	struct _btree ** next;          //pointers to the nodes children(s). If this is leaf this fieild will be NULL.
	char node_type;                 //if it's file offset is type 1 and if its descriptor then type is 2.
}btree;

typedef struct _tuple64{
	//tuples for btree node
	unsigned __int64 i_id;
	unsigned __int64 backpointer;
	unsigned __int64 offset;
}tuple64;

//descriptor leaf has different structure then other nodes.
typedef struct _tuple_leaf_d64{
	unsigned __int64 D_ID;
	unsigned __int64 DESC_I_ID;
	unsigned __int64 TREE_I_ID;
	unsigned int PARENT_D_ID;
}tuple_leaf_d64;

typedef struct _btree64{
	int num_nodes;         //number of keys in the current node.
	union /*nodes*/
	{
		tuple64 * key;          //array of structs representing all keys in a given btree
		tuple_leaf_d64 * leaf_d;
	};
	int is_leaf;           //int if 0 means this node is a leaf, otherwise if internal node, node with highest index is root.
	struct _btree64 ** next;          //pointers to the nodes children(s). If this is leaf this fieild will be NULL.
	char node_type;                 //if it's file offset is type 1 and if its descriptor then type is 2.
}btree64;

btree64* create_head (TCHAR * buffer, int num_keys, char flag);
btree64* add_node(TCHAR * buffer, int num_keys, char flag);

//internal functions
tuple64* populate_keys(unsigned __int64 * keys, unsigned __int64 *  backpointers, unsigned __int64 * offsets, int num_nodes, int num_keys);
BOOL populate1(TCHAR * buffer, btree64 * node, int num_keys, char flag);
BOOL populate2(TCHAR * buffer, btree64 * node, unsigned int items, unsigned int item_size, int num_keys, char flag);
BOOL populate3(TCHAR * buffer, btree64 * node, unsigned int items, unsigned int item_size, char flag);
void release_bbt(btree64 * head);
btree64 * get_node (btree64 * tree, unsigned __int64 i_id, int * index);

#endif